#include "StdAfx.h"
#include "GraffUnit.h"
#include "ProgressBar.h"
#include "fftw3.h"
#include "math.h"


#define MIN(a, b)  (((a) < (b)) ? (a) : (b))
#define MAX(a, b)  (((a) > (b)) ? (a) : (b))
#define Pi 3.1415926

static int sign(int i)
{
   if(i == 0) return 0;
   if(i  > 0) return 1;
   return -1;
}


CGraffUnit::CGraffUnit(CImage * pImage)
{
	m_pImage = pImage;
	work = false;
	stop = false;
	UseZBuffer = false;
}

CGraffUnit::~CGraffUnit(void)
{
}

bool CGraffUnit::Show(CDC * pDC, float ZoomFactor, CRect position)
{
	int width, height;
	int x0, y0, x1, y1;
	
	width = position.right - position.left - 3 ;// - 20;						//width of the window
	height = position.bottom - position.top - 3;// - 40;
	
	x0 = (width - (int) (ZoomFactor * m_pImage->GetWidth())) / 2;	//startig image position in the window
	y0 = (height - (int) (ZoomFactor * m_pImage->GetHeight())) / 2;
	x0 = MAX(0, x0);
	y0 = MAX(0, y0);
	
	
	x1 = x0 + (int) (ZoomFactor * m_pImage->GetWidth());
	y1 = y0 + (int) (ZoomFactor * m_pImage->GetHeight());

	

	//pDC->FillRect(CRect(0, 0, width, height), &CBrush(RGB(100, 100, 100)));

	pDC->FillRect(CRect(0, 0, width, y0), &CBrush(RGB(100, 100, 100)));
	pDC->FillRect(CRect(0, y1, width, height), &CBrush(RGB(100, 100, 100)));
	pDC->FillRect(CRect(0, y0, x0, y1), &CBrush(RGB(100, 100, 100)));
	pDC->FillRect(CRect(x1, y0, width, y1), &CBrush(RGB(100, 100, 100)));
	


	SetStretchBltMode(pDC->GetSafeHdc(), HALFTONE);
	if (UseZBuffer)
		zBuffer.Draw(pDC->GetSafeHdc(), 
			CRect(x0, y0, x0 + (int) (ZoomFactor * zBuffer.GetWidth()), y0 + (int) (ZoomFactor * zBuffer.GetHeight())), 
			CRect(0, 0, zBuffer.GetWidth(), zBuffer.GetHeight()));
	else
		m_pImage->Draw(pDC->GetSafeHdc(), 
			CRect(x0, y0, x0 + (int) (ZoomFactor * m_pImage->GetWidth()), y0 + (int) (ZoomFactor * m_pImage->GetHeight())), 
			CRect(0, 0, m_pImage->GetWidth(), m_pImage->GetHeight()));

	return true;
}

bool CGraffUnit::CreateZBuffer()
{
	if (zBuffer != NULL) zBuffer.Destroy();
	zBuffer.Create(m_pImage->GetWidth(), m_pImage->GetHeight(), m_pImage->GetBPP());
	return true;
}

void CGraffUnit::BlurAverage(void)
{
	// ---  EXAMPLE CODE  ---
	BYTE *c;
	int pitch = m_pImage->GetPitch();
	int x, y;
	long R = 0, G = 0, B = 0;
	
	c = (BYTE *) m_pImage->GetBits();
	
	for (x = 0; x < m_pImage->GetWidth(); x++)
		for (y = 0; y < m_pImage->GetHeight(); y++) {
			B += c[3*x + pitch*y + 0];
			G += c[3*x + pitch*y + 1];
			R += c[3*x + pitch*y + 2];
		}
	R /= (m_pImage->GetWidth()*m_pImage->GetHeight());
	G /= (m_pImage->GetWidth()*m_pImage->GetHeight());
	B /= (m_pImage->GetWidth()*m_pImage->GetHeight());
	for (x = 0; x < m_pImage->GetWidth(); x++)
		for (y = 0; y < m_pImage->GetHeight(); y++) {
			c[3*x + pitch*y + 0] = (BYTE) B;
			c[3*x + pitch*y + 1] = (BYTE) G;
			c[3*x + pitch*y + 2] = (BYTE) R;
		}
}


void CGraffUnit::BlurMotion(double sigma, int angle) {}
void CGraffUnit::BlurGaussian(float sigma) {}
void CGraffUnit::AddNoise(float sigma, bool distr, bool mono){}
void CGraffUnit::Invert(void)
{
	// ---  EXAMPLE CODE  ---
	BYTE	*c;
	int		pitch = m_pImage->GetPitch();
	register int x, y;
	BYTE	R, G, B;

	c = (BYTE *) m_pImage->GetBits();

	for (x = 0; x < 3 * m_pImage->GetWidth(); x += 3)
		for (y = 0; y < m_pImage->GetHeight(); y++) {
			B = c[x + pitch*y + 0];
			G = c[x + pitch*y + 1];
			R = c[x + pitch*y + 2];

			c[x + pitch*y + 0] = 255 - B;
			c[x + pitch*y + 1] = 255 - G;
			c[x + pitch*y + 2] = 255 - R;
		}
}

void CGraffUnit::Rotate180(void)
{
	// ---  EXAMPLE CODE  ---
	BYTE	*c, tmp;
	int		pitch = m_pImage->GetPitch();
	int		width = m_pImage->GetWidth(), height = m_pImage->GetHeight();
	int		halfHeight = height / 2, halfWidth = width / 2;
	int		x, y;

	c = (BYTE *) m_pImage->GetBits();

	for (y = 0; y < halfHeight; y++)
		for (x = 0; x < width; x++) {
			tmp = c[3*x + pitch*y + 0];
			c[3*x + pitch*y + 0] = c[3*(width - 1 - x) + pitch*(height - 1 - y) + 0];
			c[3*(width - 1 - x) + pitch*(height - 1 - y) + 0] = tmp;
			tmp = c[3*x + pitch*y + 1];
			c[3*x + pitch*y + 1] = c[3*(width - 1 - x) + pitch*(height - 1 - y) + 1];
			c[3*(width - 1 - x) + pitch*(height - 1 - y) + 1] = tmp;
			tmp = c[3*x + pitch*y + 2];
			c[3*x + pitch*y + 2] = c[3*(width - 1 - x) + pitch*(height - 1 - y) + 2];
			c[3*(width - 1 - x) + pitch*(height - 1 - y) + 2] = tmp;
		}
	
	if ((height % 2) != 0) {
		y = halfHeight;
		for (x = 0; x < halfWidth; x++) {
			tmp = c[3*x + pitch*y + 0];
			c[3*x + pitch*y + 0] = c[3*(width - 1 - x) + pitch*y + 0];
			c[3*(width - 1 - x) + pitch*y + 0] = tmp;
			tmp = c[3*x + pitch*y + 1];
			c[3*x + pitch*y + 1] = c[3*(width - 1 - x) + pitch*y + 1];
			c[3*(width - 1 - x) + pitch*y + 1] = tmp;
			tmp = c[3*x + pitch*y + 2];
			c[3*x + pitch*y + 2] = c[3*(width - 1 - x) + pitch*y + 2];
			c[3*(width - 1 - x) + pitch*y + 2] = tmp;
		}
	}

}

void CGraffUnit::Rotate90CW(void)
{
	// ---  EXAMPLE CODE  ---
	CImage	tmp;
	BYTE	*c, *c_tmp;
	int		pitch, pitch_tmp;
	int		width, height;
	int		x, y;

	tmp.Create(m_pImage->GetWidth(), m_pImage->GetHeight(), m_pImage->GetBPP());
	m_pImage->BitBlt(tmp.GetDC(), 0, 0, SRCCOPY);
	tmp.ReleaseDC();
	m_pImage->Destroy();
	m_pImage->Create(tmp.GetHeight(), tmp.GetWidth(), tmp.GetBPP());
	
	c = (BYTE *) m_pImage->GetBits();
	pitch = m_pImage->GetPitch();
	c_tmp = (BYTE *) tmp.GetBits();
	pitch_tmp = tmp.GetPitch();

	width = m_pImage->GetWidth();
	height = m_pImage->GetHeight();

	for (x = 0; x < width; x++)
		for (y = 0; y < height; y++){
			c[3*x + pitch*y + 0] = c_tmp[3*y + pitch_tmp*(width - 1 - x) + 0];
			c[3*x + pitch*y + 1] = c_tmp[3*y + pitch_tmp*(width - 1 - x) + 1];
			c[3*x + pitch*y + 2] = c_tmp[3*y + pitch_tmp*(width - 1 - x) + 2];
		}
}

void CGraffUnit::Rotate90CCW(void)
{
	// ---  EXAMPLE CODE  ---
	CImage	tmp;
	BYTE	*c, *c_tmp;
	int		pitch, pitch_tmp;
	int		width, height;
	int		x, y;

	tmp.Create(m_pImage->GetWidth(), m_pImage->GetHeight(), m_pImage->GetBPP());
	m_pImage->BitBlt(tmp.GetDC(), 0, 0, SRCCOPY);
	tmp.ReleaseDC();
	m_pImage->Destroy();
	m_pImage->Create(tmp.GetHeight(), tmp.GetWidth(), tmp.GetBPP());
	
	c = (BYTE *) m_pImage->GetBits();
	pitch = m_pImage->GetPitch();
	c_tmp = (BYTE *) tmp.GetBits();
	pitch_tmp = tmp.GetPitch();

	width = m_pImage->GetWidth();
	height = m_pImage->GetHeight();

	for (x = 0; x < width; x++)
		for (y = 0; y < height; y++){
			c[3*x + pitch*y + 0] = c_tmp[3*(height - 1 - y) + pitch_tmp*x + 0];
			c[3*x + pitch*y + 1] = c_tmp[3*(height - 1 - y) + pitch_tmp*x + 1];
			c[3*x + pitch*y + 2] = c_tmp[3*(height - 1 - y) + pitch_tmp*x + 2];
		}
}

void CGraffUnit::FlipHorizontal(void)
{
	// ---  EXAMPLE CODE  ---
	BYTE	*c, tmp;
	int		pitch = m_pImage->GetPitch();
	int		width = m_pImage->GetWidth(), height = m_pImage->GetHeight();
	int		halfHeight = height / 2, halfWidth = width / 2;
	int		x, y;

	c = (BYTE *) m_pImage->GetBits();

	for (x = 0; x < halfWidth; x++)
		for (y = 0; y < height; y++) {
			tmp = c[3*x + pitch*y + 0];
			c[3*x + pitch*y + 0] = c[3*(width - 1 - x) + pitch*y + 0];
			c[3*(width - 1 - x) + pitch*y + 0] = tmp;
			tmp = c[3*x + pitch*y + 1];
			c[3*x + pitch*y + 1] = c[3*(width - 1 - x) + pitch*y + 1];
			c[3*(width - 1 - x) + pitch*y + 1] = tmp;
			tmp = c[3*x + pitch*y + 2];
			c[3*x + pitch*y + 2] = c[3*(width - 1 - x) + pitch*y + 2];
			c[3*(width - 1 - x) + pitch*y + 2] = tmp;
		}
}

void CGraffUnit::FlipVertical(void)
{
	// ---  EXAMPLE CODE  ---
	BYTE	*c, tmp;
	int		pitch = m_pImage->GetPitch();
	int		width = m_pImage->GetWidth(), height = m_pImage->GetHeight();
	int		halfHeight = height / 2, halfWidth = width / 2;
	int		x, y;

	c = (BYTE *) m_pImage->GetBits();

	for (y = 0; y < halfHeight; y++)
		for (x = 0; x < width; x++) {
			tmp = c[3*x + pitch*y + 0];
			c[3*x + pitch*y + 0] = c[3*x + pitch*(height - 1 - y) + 0];
			c[3*x + pitch*(height - 1 - y) + 0] = tmp;
			tmp = c[3*x + pitch*y + 1];
			c[3*x + pitch*y + 1] = c[3*x + pitch*(height - 1 - y) + 1];
			c[3*x + pitch*(height - 1 - y) + 1] = tmp;
			tmp = c[3*x + pitch*y + 2];
			c[3*x + pitch*y + 2] = c[3*x + pitch*(height - 1 - y) + 2];
			c[3*x + pitch*(height - 1 - y) + 2] = tmp;
		}
}

void CGraffUnit::Desaturate(void) {}
void CGraffUnit::Affine(float a, float b){}
void CGraffUnit::Resize(int X, int Y, char resample)
{
	// ---  EXAMPLE CODE  ---
	CImage	tmp;
	BYTE	*c, *c_tmp;
	int		pitch, pitch_tmp;
	int		width, height;
	int		x, y;
	float	dx, dy;
	
	tmp.Create(m_pImage->GetWidth(), m_pImage->GetHeight(), m_pImage->GetBPP());
	m_pImage->BitBlt(tmp.GetDC(), 0, 0, SRCCOPY);
	tmp.ReleaseDC();

	m_pImage->Destroy();
	m_pImage->Create(X, Y, tmp.GetBPP());

	c = (BYTE *) m_pImage->GetBits();
	pitch = m_pImage->GetPitch();
	c_tmp = (BYTE *) tmp.GetBits();
	pitch_tmp = tmp.GetPitch();

	width = tmp.GetWidth();
	height = tmp.GetHeight();

	dx = (float) width / (float) X;
	dy = (float) height / (float) Y;

	switch (resample) {
		case 'n':
		//Nearest Neightbor Method
		for (x = 0; x < X; x++)
			for (y = 0; y < Y; y++) {
				c[3*x + pitch*y + 0] = c_tmp[3* (int) (dx*x) + pitch_tmp * (int) (dx*y) + 0];
				c[3*x + pitch*y + 1] = c_tmp[3* (int) (dx*x) + pitch_tmp * (int) (dx*y) + 1];
				c[3*x + pitch*y + 2] = c_tmp[3* (int) (dx*x) + pitch_tmp * (int) (dx*y) + 2];
			}
		break;
	
		case 'b':
		//Bilenear filtering
	
		SetStretchBltMode(m_pImage->GetDC(), HALFTONE);
		tmp.Draw(m_pImage->GetDC(), CRect(0, 0, X, Y), CRect(0, 0, width, height));
		m_pImage->ReleaseDC();
		m_pImage->ReleaseDC();
		break;
	}

}

void CGraffUnit::Equalization(void){}
void CGraffUnit::Median(int Radius){}
void CGraffUnit::Sharpen(void)
{
	// ---  EXAMPLE CODE  ---
	CImage	tmp;
	BYTE	*c, *c_tmp;
	int		pitch = m_pImage->GetPitch(), pitch_tmp;
	int		x, y;
	long	C;

	tmp.Create(m_pImage->GetWidth(), m_pImage->GetHeight(), m_pImage->GetBPP());
	m_pImage->BitBlt(tmp.GetDC(), 0, 0, SRCCOPY);
	tmp.ReleaseDC();

	c = (BYTE *) m_pImage->GetBits();
	c_tmp = (BYTE *) tmp.GetBits();
	pitch_tmp = tmp.GetPitch();

	for (x = 1; x < m_pImage->GetWidth() - 1; x++)
		for (y = 1; y < m_pImage->GetHeight() - 1; y++) {
			C = 5 * c_tmp[3*x + pitch_tmp*y + 0] - c_tmp[3*(x+1) + pitch_tmp*y + 0] - c_tmp[3*(x-1) + pitch_tmp*y + 0] - c_tmp[3*x + pitch_tmp*(y-1) + 0] - c_tmp[3*x + pitch_tmp*(y+1) + 0];
			if (C > 255) C = 255; if (C < 0) C = 0; 
			c[3*x + pitch*y + 0] = (BYTE) C;
			C = 5 * c_tmp[3*x + pitch_tmp*y + 1] - c_tmp[3*(x+1) + pitch_tmp*y + 1] - c_tmp[3*(x-1) + pitch_tmp*y + 1] - c_tmp[3*x + pitch_tmp*(y-1) + 1] - c_tmp[3*x + pitch_tmp*(y+1) + 1];
			if (C > 255) C = 255; if (C < 0) C = 0; 
			c[3*x + pitch*y + 1] = (BYTE) C;
			C = 5 * c_tmp[3*x + pitch_tmp*y + 2] - c_tmp[3*(x+1) + pitch_tmp*y + 2] - c_tmp[3*(x-1) + pitch_tmp*y + 2] - c_tmp[3*x + pitch_tmp*(y-1) + 2] - c_tmp[3*x + pitch_tmp*(y+1) + 2];
			if (C > 255) C = 255; if (C < 0) C = 0; 
			c[3*x + pitch*y + 2] = (BYTE) C;
		}
}


UINT CGraffUnit::TotalVariation(LPVOID lParam)
{
	// ---  DUMMY CODE  ---
	UGU_STRUCT * UGUStructure = (UGU_STRUCT *) lParam;
	CGraffUnit * me = (CGraffUnit *) UGUStructure->pVoid;
	me->work = true;
	me->stop = false;
	int N = UGUStructure->N;

	me->work = false;
	return 0;
}

UINT CGraffUnit::BalancedForwardBackward(LPVOID lParam)
{
	// ---  DUMMY CODE  ---
	UGU_STRUCT * UGUStructure = (UGU_STRUCT *) lParam;
	CGraffUnit * me = (CGraffUnit *) UGUStructure->pVoid;
	me->work = true;
	me->stop = false;

	int N = UGUStructure->N;

	me->work = false;
	return 0;
}

void CGraffUnit::AutoColor(void) {}
void CGraffUnit::RedEye(void) {}